#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes the Agreement Rate (AR) measure from an agreement matrix A.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
AR <- function(A) 
{
  N <- nrow(A)
  return ((sum(A)-N)/(N*(N-1))) # subtract N from the sum because A[i,i] is 1 for any agreement matrix
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes the Disagreement Rate (DR) measure from an agreement matrix A.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
DR <- function(A) 
{
  return (1-AR(A))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes the coagreement rate (CR) measure from two agreement matrixes A1 and A2.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
CR <- function(A1, A2) 
{
  N <- nrow(A1)
  return ((sum(A1*A2)-N)/(N*(N-1))) # subtract N from the sum because A[i,i] is 1 for any coagreement matrix
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes individual agreement rates (per user) from the agreement matrix A.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
ARs <- function(A) 
{
  N <- nrow(A)
  return ((rowSums(A)-1)/(N-1)) # subtract 1 from each row sum since A[i,i]=1
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Variations of the Agreement Rate (AR) measure from an agreement matrix A.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
AR.mean <- function(A) 
{ 
  return (mean(ARs(A))) 
}

AR.median <- function(A) 
{ 
  return (median(ARs(A))) 
}

AR.tmean <- function(A, trim = 0.2) 
{ 
  return (tmean(ARs(A), tr = trim)) 
}

AR.winmean <- function(A, trim = 0.2) 
{ 
  return (winmean(ARs(A), tr = trim)) 
}

AR.family <- function(A) 
{
  return (list(
    Mean = AR.mean(A),
    Median = AR.median(A),
    TrimmedMean = AR.tmean(A),
    WinMean = AR.winmean(A)
  ))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes an agreement matrix from a dissimilarity matrix (DM)
# by applying the tolerance epsilon.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
AM.fromDM <- function(DM, epsilon) 
{
  return ((DM <= epsilon) * 1)
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes the tolerances automatically using the quantiles of the 
# values from the dissimilarity matrix (DM) so that the agreement rates 
# are exactly TARGET_ARs (or close to TARGET_ARs for low N values).
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Epsilon.fromDM.ExactARs <- function(DM, TARGET_ARs)
{
  # convert the dissimilarity matrix into a vector and remove the zero values from the main diagonal
  N <- nrow(DM)
  v <- sort(as.vector(DM))
  v <- v[c((N+1):length(v))] # the first N values are zeros from the main diagonal of the dissimilarity matrix
  
  # compute the quantiles corresponding to the target agreement rate
  q <- quantile(v, probs = TARGET_ARs)
  
  # use the quantiles as the tolerance epsilon values
  return (Epsilon = as.vector(q))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes an agreement matrix from a set of proposals elicited from end users.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
AM.fromProposals <- function(proposals) 
{
  N <- length(proposals)
  AM <- matrix(rep(0, N*N), nrow = N, ncol = N)
  for(i in 1:N)
    for(j in i:N)
    {
      AM[i,j] <- ifelse(proposals[i] == proposals[j], 1, 0)
      AM[j,i] <- AM[i,j]
    }
  return (AM)
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Computes the non-transitivity rate (NTR) measures for an agreement matrix.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NTR <- function(A) 
{
  N <- nrow(A)
  
  t <- 0     # number of transitive relations
  nt <- 0    # number of non-transitive relations
  count <- 0 # all relations
  
  for(i in 1:(N-2))
    for(j in (i+1):(N-1))
      for(k in (j+1):N)
      {
        sum <- A[i,j] + A[j,k] + A[i,k]
        if (sum == 3) t <- t + 1
        if (sum == 2) nt <- nt + 1
        # if (sum == 1) no conclusion
        # if (sum == 0) no conclusion
        count <- count + 1
      }
  
  return (list(
    NumRelations = count,
    Transitive = t,
    NonTransitive = nt,
    NTR1 = ifelse(count < 1, NaN, nt/count * 100), 
    NTR2 = ifelse(t+nt < 1, NaN, nt/(t+nt) * 100)
  ))
}