#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the Manhattan distance and scalars
# sampled from a normal distribution with mean = params[1] and sd = params[2].
# Returns both the dissimilarity matrix and the values from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.Simple <- function(N, params = c(0,1)) 
{
  # sample N values from the normal distribution
  values <- rnorm(N, mean = params[[1]], sd = params[[2]])
  
  # compute the dissimilarity matrix
  # using the Manhattan distance
  DM <- matrix(nrow = N, ncol = N)
  for(i in 1:N) 
    for(j in i:N) {
      DM[i,j] <- abs(values[i] - values[j])
      DM[j,i] <- DM[i,j]
    }
  return (list(DM = DM, values = values))
}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the squared Euclidean distance and p-dimensional
# feature vectors (i.e., proposals) with the values of the p features sampled from normal
# distributions with means 0 and SDs specified by the params array.
# The proposals can be either generated inside the function or provided as an argument.
# Returns the dissimilarity matrix and the vectors (proposals) from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.SqrEuclidean <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE)
{
  # number of dimensions to characterize users' proposals
  k <- length(params)
  
  # generate users' proposals (or use the proposals provided in the list of arguments)
  if (is.null(proposals))
  {
    users <- matrix(nrow = N, ncol = k)
    for(i in 1:k)
      users[,i] <- rnorm(N, mean = 0, sd = params[[i]])
  }
  else users <- proposals
  
  # sometimes, the list of proposals is all that's needed from this function
  if (GENERATE_PROPOSALS_ONLY)
    return (list(DM = NULL, values = users))
  
  # compute the dissimilarity matrix
  DM <- matrix(nrow = N, ncol = N)
  for(i in 1:N) 
    for(j in i:N)
    {
      DM[i,j] <- dist(rbind(users[i,], users[j,]))^2
      DM[j,i] <- DM[i,j]
    }
  
  return (list(DM = DM, values = users))
}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the Minkowski distance and p-dimensional
# feature vectors (i.e., proposals) with the values of the p features sampled from normal distributions
# with means 0 and SDs specified by the params array.
# The proposals can be either generated inside the function or provided as an argument.
# Returns the dissimilarity matrix and the vectors (proposals) from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.Minkowski <- function(N, params, p, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE)
{
  # number of dimensions to characterize users' proposals
  k <- length(params) 
  
  # generate users' proposals (or use the proposals provided in the list of arguments)
  if (is.null(proposals))
  {
    users <- matrix(nrow = N, ncol = k)
    for(i in 1:k)
      users[,i] <- rnorm(N, mean = 0, sd = params[[i]])
  }
  else users <- proposals
  
  # sometimes, the list of proposals is all that's needed from this function
  if (GENERATE_PROPOSALS_ONLY)
    return (list(DM = NULL, values = users))
  
  # compute dissimilarity matrix
  DM <- suppressMessages(philentropy::distance(users, method = "minkowski", p))
  return (list(DM = DM, values = users))
}

generateDM.Minkowski.0.5 <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE) {
  return (generateDM.Minkowski(N, params, 0.5, proposals, GENERATE_PROPOSALS_ONLY))
}
generateDM.Minkowski.1 <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE) {
  return (generateDM.Minkowski(N, params, 1, proposals, GENERATE_PROPOSALS_ONLY))
}
generateDM.Minkowski.1.5 <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE) {
  return (generateDM.Minkowski(N, params, 1.5, proposals, GENERATE_PROPOSALS_ONLY))
}
generateDM.Minkowski.2 <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE) {
  return (generateDM.Minkowski(N, params, 2, proposals, GENERATE_PROPOSALS_ONLY))
}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the overlap distance and vectors of 
# categories (i.e., proposals) sampled from Zipf-Mandelbrot distributions with B's specified by the params array.
# The proposals can be either generated inside the function or provided as an argument.
# Returns the dissimilarity matrix and the vectors (proposals) from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.ZipfMandelbrot.Overlap <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE)
{
  # number of dimensions to characterize users' proposals
  k <- length(params) 
  
  # generate users' proposals (or use the proposals provided in the list of arguments)
  if (is.null(proposals))
  {
    users <- matrix(nrow = N, ncol = k)
    for(i in 1:k)
      users[,i] <- as.character(rlnre(lnre("zm", alpha = 1/2, B = params[[i]]), n = N))
  }
  else users <- proposals
  
  # sometimes, the list of proposals is all that's needed from this function
  if (GENERATE_PROPOSALS_ONLY)
    return (list(DM = NULL, values = users))
  
  # compute dissimilarity matrix
  DM <- matrix(nrow = N, ncol = N)
  for(i in 1:N) 
    for(j in i:N) 
    {
      DM[i,j] <- overlap.distance(users[i,], users[j,])
      DM[j,i] <- DM[i,j]
    }
  return (list(DM = DM, values = users))
}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the Lin distance and vectors of 
# categories (i.e., proposals) sampled from Zipf-Mandelbrot distributions with B's specified by the params array.
# The proposals can be either generated inside the function or provided as an argument.
# Returns the dissimilarity matrix and the vectors from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.ZipfMandelbrot.Lin <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE)
{
  # number of dimensions to characterize users' proposals
  k <- length(params) 
  
  ZMmodels <- list()
  for(i in 1:k)
    ZMmodels[[i]] <- lnre("zm", alpha = 1/2, B = params[[i]])
    
  # generate users' proposals (or use the proposals provided in the list of arguments)
  if (is.null(proposals))
  {
    users <- matrix(nrow = N, ncol = k)
    for(i in 1:k)
      users[,i] <- as.character(rlnre(ZMmodels[[i]], n = N))
  }
  else users <- proposals
    
  # sometimes, the list of proposals is all that's needed from this function
  if (GENERATE_PROPOSALS_ONLY)
    return (list(DM = NULL, values = users))
  
  # compute dissimilarity matrix
  DM <- matrix(nrow = N, ncol = N)
  for(i in 1:N) 
    for(j in i:N) 
    {
      DM[i,j] <- Lin.distance.ZM(users[i,], users[j,], ZMmodels)
      DM[j,i] <- DM[i,j]
    }
  
  # the Lin distance returns negative values, so
  # subtract min(DM) from all the values, which will make DM positively defined
  DM <- DM - min(DM)
  for(i in 1:N) DM[i,i] <- 0
  
  return (list(DM = DM, values = users))
}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Generates a N x N dissimilarity matrix for the overlap distance (Lin) and 
# binary vectors (i.e., proposals) sampled from Bernoulli distributions with probabilities of success
# specified by the params array. 
# The proposals can be either generated inside the function or provided as an argument.
# Returns  the dissimilarity matrix and the vectors (proposals) from which it was computed.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
generateDM.Bernoulli.Lin <- function(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = FALSE)
{
  # number of dimensions on which to characterize users' proposals
  k <- length(params)
  
  # generate users' proposals (or use the proposals provided in the list of arguments)
  if (is.null(proposals))
  {
    users <- matrix(nrow = N, ncol = k)
    for(i in 1:k)
      users[,i] <- rbinom(N, 1, params[[i]])
  }
  else users <- proposals
  
  # sometimes, the list of proposals is all that's needed from this function
  if (GENERATE_PROPOSALS_ONLY)
    return (list(DM = NULL, values = users))
  
  # compute dissimilarity matrix
  DM <- matrix(nrow = N, ncol = N)
  for(i in 1:N)
    for(j in i:N)
    {
      DM[i,j] <- Lin.distance.Bernoulli(users[i,], users[j,], params)
      DM[j,i] <- DM[i,j]
    }
  
  # Lin's distance returns negative values, but by subtracting min(DM), DM will become positively defined
  DM <- DM - min(DM)
  for(i in 1:N) DM[i,i] <- 0
  
  return (list(DM = DM, values = users))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Implements the "overlap" distance (Stanfill and Waltz, 1986) [p. 1219] as the number
# of elements for which two sets, a and b, have different values; e.g., the overlap distance 
# between sets (1,3,4,5,7) and c(1,0,4,2,3) is 3/5=0.6, because the pairs 3-0, 5-2, 
# and 7-3 do not match. Although the overlap distance is not the best measure for comparing
# categorical data, it is remarkably simple and fares rather well compared to other, 
# more complex measures; see Boriah et al. (2008) for comparative evaluations.
#
# The overlap distance takes values between 0 and 1.
#
# C. Stanfill, D. Waltz. 1986. Toward memory-based reasoning. Commun. ACM 29, 12, 1213-1228
# http://dx.doi.org/10.1145/7902.7906
#
# S. Boriah, V. Chandola, V. Kumar. 2008. Similarity measures for categorical data: 
# A comparative evaluation. In Proc. of the 8th SIAM Int. Conference on Data Mining 
# https://pdfs.semanticscholar.org/c654/4a12fec2097bddc49adbb159426d9dc15d2c.pdf
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
overlap.distance <- function(a, b)
{
  match <- a == b
  return (length(match[match == FALSE])/length(match))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Implements the Lin distance, described in Boriah et al. (2008).
# According to the evaluations from Boriah et al. (2008), the Lin distance
# delivered the best performance accross several datasets.
#
# - a is a random sample, for which the probability to observe attribute a[i] is proba[i]
# - b is a random sample, for which the probability to observe attribute b[i] is probb[i]
#
# S. Boriah, V. Chandola, V. Kumar. 2008. Similarity measures for categorical data: 
# A comparative evaluation. In Proc. of the 8th SIAM Int. Conference on Data Mining 
# https://pdfs.semanticscholar.org/c654/4a12fec2097bddc49adbb159426d9dc15d2c.pdf
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Lin.distance <- function(a, b, proba, probb) {
  N <- length(a)
  w <- 1
  S <- 1
  for(i in 1:N) {
    S <- S * ifelse(a[[i]] == b[[i]], proba[[i]], proba[[i]] + probb[[i]])
    w <- w * proba[[i]] * probb[[i]]
  }
  S <- 2 * log(S)/log(w)
  return (-S) # convert a similiarity measure into a dissimilarity
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Implements the Lin distance for Bernoulli trials.
#
# - a is an array of Bernoulli trials with probability of success p[i] for the i-th attribute
# - b is an array of Bernoulli trials with probability of success p[i] for the i-th attribute
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Lin.distance.Bernoulli <- function(a, b, p)
{
  N <- length(a)
  proba <- p
  probb <- p
  for(i in 1:N) {
    if (a[[i]] == 0) proba[[i]] <- 1 - p[[i]]
    if (b[[i]] == 0) probb[[i]] <- 1 - p[[i]]
  }
  return (Lin.distance(a, b, proba, probb))
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Implements the Lin distance for Zipf-Mandelbrot (ZM) trials.
#
# - a is a random sample from a ZM distribution
# - b is a random sample from a ZM distribution
# - ZMmodels represents the ZM distributions for each feature i
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Lin.distance.ZM <- function(a, b, ZMmodels)
{
  probZM <- function(ZMmodel, k) {
    return (plnre(ZMmodel, tqlnre(ZMmodel, k-1)) - plnre(ZMmodel, tqlnre(ZMmodel, k)))
  }
  
  a <- as.numeric(as.character(a))
  b <- as.numeric(as.character(b))
  N <- length(a)
  proba <- rep(0,N)
  probb <- rep(0,N)
  for(i in 1:N) {
    proba[[i]] <- probZM(ZMmodels[[i]], a[[i]])
    probb[[i]] <- probZM(ZMmodels[[i]], b[[i]])
  }
  return (Lin.distance(a, b, proba, probb))
}