#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Runs unit tests for agreement measures, dissimilariy functions, simulations, etc.
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
runAllUnitTests <- function()
{
  ## unit tests for agreement rates
  unittests.AR()
  unittests.AM.fromProposals()
  unittests.NTR()
  unittests.AM.fromDM()
  
  ## unit tests for dissimilarity functions
  unittests.overlap.distance()
  unittests.Lin.distance()
  unittests.Lin.distance.Bernoulli()
  unittests.Lin.distance.ZM()
  
  ## unit tests for simulations
  unittests.generateDM.Simple()
  unittests.generateDM.SqrEuclidean()
  unittests.generateDM.Minkowski()
  unittests.generateDM.ZipfMandelbrot.Overlap()
  unittests.generateDM.ZipfMandelbrot.Lin()
  unittests.generateDM.Bernoulli.Lin()
  
  SDs <- c(5.42, 2.56, 1.63, 1.15, .88, .69, .576, .493, .416)  # SD parameters for the normal distributions (adopted from T. Tsandilas)
  betas <- c(.306, .64, 1.05, 1.56, 2.25, 3.26, 5.0, 8.3, 18.4) # B parameters for the Zipf-Mandelbrot distributions (adopted from T. Tsandilas)
  NUM_TRIALS <- 1
  unittests.generateDM(FUN = generateDM.SqrEuclidean, params = c(0, 1), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = FALSE, VERBOSE = TRUE)
  for(k in 1:10) {
    cat("Running tests for population generation (number of dimensions k =", k, ")\r\n")
    
    #cat("Minkowski...\r\n")
    unittests.generateDM(FUN = generateDM.Minkowski.0.5, params = sample(SDs, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = ifelse(k==1, TRUE, FALSE))
    unittests.generateDM(FUN = generateDM.Minkowski.1,   params = sample(SDs, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = ifelse(k==1, TRUE, FALSE))
    unittests.generateDM(FUN = generateDM.Minkowski.1.5, params = sample(SDs, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = ifelse(k==1, TRUE, FALSE))
    unittests.generateDM(FUN = generateDM.Minkowski.2,   params = sample(SDs, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = TRUE)
    
    #cat("Overlap...\r\n")
    unittests.generateDM(FUN = generateDM.ZipfMandelbrot.Overlap, params = sample(betas, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = FALSE)
    
    #cat("Lin...\r\n")
    unittests.generateDM(FUN = generateDM.ZipfMandelbrot.Lin,     params = sample(betas, k, replace = TRUE), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = FALSE)
    unittests.generateDM(FUN = generateDM.Bernoulli.Lin, params = runif(k, 0, 1), TRIALS = NUM_TRIALS, TRIANGLE_INEQUALITY = FALSE)
  }
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for Agreement Rates
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.AR <- function() 
{
  #-----------------------------------
  # test: AR is 0 for null agreement graphs
  #-----------------------------------
  cat("Running AR = 0 tests ... ")
  numTests <- 0
  for(N in (2:100))
  {
    A <- matrix(data = rep(0, N*N), nrow = N, ncol = N)
    for(i in 1:N) A[i,i] <- 1
    
    expect_equal(AR(A), 0, tolerance = 0)
    expect_equal(DR(A), 1, tolerance = 0)
    expect_equal(AR.mean(A), 0, tolerance = 0)
    expect_equal(AR.family(A)$Mean, 0, tolerance = 0)
    expect_equal(AR.median(A), 0, tolerance = 0)
    expect_equal(AR.family(A)$Median, 0, tolerance = 0)
    expect_equal(AR.tmean(A), 0, tolerance = 0)
    expect_equal(AR.family(A)$TrimmedMean, 0, tolerance = 0)
    expect_equal(AR.winmean(A), 0, tolerance = 0)
    expect_equal(AR.family(A)$WinMean, 0, tolerance = 0)
    numTests <- numTests + 1
  }
  cat(numTests, "tests done.\r\n")
  
  #-----------------------------------
  # test: AR is 1 for complete agreement graphs
  #-----------------------------------
  cat("Running AR = 1 tests ... ")
  numTests <- 0
  for(N in (2:100))
  {
    A <- matrix(data = rep(1, N*N), nrow = N, ncol = N)
    expect_equal(AR(A), 1, tolerance = 0)
    expect_equal(DR(A), 0, tolerance = 0)
    expect_equal(AR.mean(A), 1, tolerance = 0)
    expect_equal(AR.family(A)$Mean, 1, tolerance = 0)
    expect_equal(AR.median(A), 1, tolerance = 0)
    expect_equal(AR.family(A)$Median, 1, tolerance = 0)
    expect_equal(AR.tmean(A), 1, tolerance = 0)
    expect_equal(AR.family(A)$TrimmedMean, 1, tolerance = 0)
    expect_equal(AR.winmean(A), 1, tolerance = 0)
    expect_equal(AR.family(A)$WinMean, 1, tolerance = 0)
    numTests <- numTests + 1
  }
  cat(numTests, "tests done.\r\n")
  
  #-----------------------------------
  # test: custom agreement rates AR
  #-----------------------------------
  cat("Running custom AR tests ... ")
  numTests <- 0
  for(N in seq(10, 100, 10))
    for(M in 1:(N-2))
    {
      A <- matrix(data = rep(0, N*N), nrow = N, ncol = N)
      for(i in 1:N) A[i,i] <- 1
      for(i in 1:M)
      {
        index <- sample(seq(i+1,N,1), 1)[[1]]
        A[i,index] <- 1
        A[index,i] <- 1
      }
      
      expect_equal(AR(A), (2*M)/(N*(N-1)), tolerance = 1e-3)
      expect_equal(DR(A), 1-(2*M)/(N*(N-1)), tolerance = 0)
      expect_equal(AR.mean(A), (2*M)/(N*(N-1)), tolerance = 1e-3)
      expect_equal(AR.family(A)$Mean, (2*M)/(N*(N-1)), tolerance = 1e-3)
      numTests <- numTests + 1
    }
  
  #-----------------------------------
  # other custom tests
  #-----------------------------------
  A <- matrix(
    data = c(1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1),
    nrow = 5, ncol = 5
  )
  expect_equal(AR(A), 0.7, tolerance = 1e-6)
  expect_equal(DR(A), 0.3, tolerance = 1e-6)
  expect_equal(AR.mean(A), 0.7, tolerance = 0)
  expect_equal(AR.family(A)$Mean, 0.7, tolerance = 0)
  expect_equal(AR.median(A), 0.75, tolerance = 0)
  expect_equal(AR.family(A)$Median, 0.75, tolerance = 0)
  expect_equal(AR.tmean(A), 0.6666667, tolerance = 1e-6)
  expect_equal(AR.family(A)$TrimmedMean, 0.6666667, tolerance = 1e-6)
  expect_equal(AR.winmean(A), 0.65, tolerance = 0)
  expect_equal(AR.family(A)$WinMean, 0.65, tolerance = 0)
  numTests <- numTests + 1
  
  A <- matrix(
    data = c(1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1),
    nrow = 7, ncol = 7
  )
  expect_equal(AR(A), 12/21, tolerance = 1e-6)
  expect_equal(DR(A), 9/21, tolerance = 1e-6)
  expect_equal(AR.mean(A), 12/21, tolerance = 1e-6)
  expect_equal(AR.family(A)$Mean, 12/21, tolerance = 1e-6)
  expect_equal(AR.median(A), 0.6666667, tolerance = 1e-6)
  expect_equal(AR.family(A)$Median, 0.6666667, tolerance = 1e-6)
  expect_equal(AR.tmean(A), 0.5666667, tolerance = 1e-6)
  expect_equal(AR.family(A)$TrimmedMean, 0.5666667, tolerance = 1e-6)
  expect_equal(AR.winmean(A), 0.547619, tolerance = 1e-6)
  expect_equal(AR.family(A)$WinMean, 0.547619, tolerance = 1e-6)
  numTests <- numTests + 1
  
  A <- matrix(
    data = c(1,1,1,1),
    nrow = 2, ncol = 2
  )
  expect_equal(AR(A), 1, tolerance = 0)
  expect_equal(DR(A), 0, tolerance = 0)
  expect_equal(AR.mean(A), 1, tolerance = 0)
  expect_equal(AR.family(A)$Mean, 1, tolerance = 0)
  expect_equal(AR.median(A), 1, tolerance = 1e-6)
  expect_equal(AR.family(A)$Median, 1, tolerance = 1e-6)
  expect_equal(AR.tmean(A), 1, tolerance = 1e-6)
  expect_equal(AR.family(A)$TrimmedMean, 1, tolerance = 1e-6)
  expect_equal(AR.winmean(A), 1, tolerance = 1e-6)
  expect_equal(AR.family(A)$WinMean, 1, tolerance = 1e-6)
  numTests <- numTests + 1
  
  A <- matrix(
    data = c(1,0,1,0,0,1,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,0,1,1,1,0,0,1,0,1,1,0,1,1,0,1,1,1,0,1,1,0,1,0,1,1,0,0,1),
    nrow = 7, ncol = 7
  )
  expect_equal(AR(A), 0.619048, tolerance = 1e-6)
  expect_equal(DR(A), 1-0.619048, tolerance = 1e-6)
  expect_equal(AR.mean(A), 0.619048, tolerance = 1e-6)
  expect_equal(AR.family(A)$Mean, 0.619048, tolerance = 1e-6)
  expect_equal(AR.median(A), 0.5, tolerance = 1e-6)
  expect_equal(AR.family(A)$Median, 0.5, tolerance = 1e-6)
  expect_equal(AR.tmean(A), 0.5666667, tolerance = 1e-6)
  expect_equal(AR.family(A)$TrimmedMean, 0.5666667, tolerance = 1e-6)
  expect_equal(AR.winmean(A), 0.5714286, tolerance = 1e-6)
  expect_equal(AR.family(A)$WinMean, 0.5714286, tolerance = 1e-6)
  numTests <- numTests + 1
  
  A <- matrix(
    data = c(1,0,1,0,1,0,1,1,1,1,1,1,1,0,0,0,1,0,1,1,1,1,0,1,1),
    nrow = 5, ncol = 5
  )
  expect_equal(AR(A), 0.6, tolerance = 1e-6)
  expect_equal(DR(A), 0.4, tolerance = 1e-6)
  expect_equal(AR.mean(A), 0.6, tolerance = 1e-6)
  expect_equal(AR.family(A)$Mean, 0.6, tolerance = 1e-6)
  expect_equal(AR.median(A), 0.5, tolerance = 1e-6)
  expect_equal(AR.family(A)$Median, 0.5, tolerance = 1e-6)
  expect_equal(AR.tmean(A), 0.5833333, tolerance = 1e-6)
  expect_equal(AR.family(A)$TrimmedMean, 0.5833333, tolerance = 1e-6)
  expect_equal(AR.winmean(A), 0.6, tolerance = 1e-6)
  expect_equal(AR.family(A)$WinMean, 0.6, tolerance = 1e-6)
  numTests <- numTests + 1
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for Agreement Rates
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.AM.fromProposals <- function() 
{
  numTests <- 0

  cat("Running AR tests...")
  
  for(N in seq(10, 100, 10)) # test with out samples of different sizes N
  {
    #-----------------------------------
    # test: zero agreement matrix
    #-----------------------------------
    A <- AM.fromProposals(sample(1:N, N, replace = FALSE))
    expect_equal(nrow(A), expected = N)
    expect_equal(ncol(A), expected = N)
    expect_true(isSymmetric(A))
    expect_true(all(unique(diag(A)) == c(1)))
    expect_equal(AR(A), expected = 0, tolerance = 0)
    expect_equal(DR(A), expected = 1, tolerance = 0)
    numTests <- numTests + 1
    
    #-----------------------------------
    # test: complete agreement matrix 
    #-----------------------------------
    A <- AM.fromProposals(rep(1, N))
    expect_equal(nrow(A), expected = N)
    expect_equal(ncol(A), expected = N)
    expect_true(isSymmetric(A))
    expect_true(all(unique(diag(A)) == c(1)))
    expect_equal(AR(A), expected = 1, tolerance = 0)
    expect_equal(DR(A), expected = 0, tolerance = 0)
    numTests <- numTests + 1
    
    #-----------------------------------
    # test: various agreement rates
    #-----------------------------------
    for(a in 1:(N-1))
    {
      A <- AM.fromProposals(c(rep(1,a),seq(1000,1000+N-a-1,1)))
      expect_equal(nrow(A), expected = N)
      expect_equal(ncol(A), expected = N)
      expect_true(isSymmetric(A))
      expect_true(all(unique(diag(A)) == c(1)))
      expect_equal(AR(A), expected = a*(a-1)/(N*(N-1)), tolerance = 0)
      expect_equal(DR(A), expected = 1 - a*(a-1)/(N*(N-1)), tolerance = 0)
      numTests <- numTests + 1
    }
  }
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for NTR
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.NTR <- function() 
{
  cat("Running NTR tests...")
  numTests <- 0
  
  #-----------------------------------
  #test: zero matrices
  #-----------------------------------
  for(N in 5:20)
  {
    A <- matrix(rep(0, N*N), nrow = N, ncol = N)
    for(i in 1:N) A[i,i] <- 1
    
    ntr <- NTR(A)
    expect_equal(ntr$NumRelations, expected = choose(N,3), tolerance = 0)
    expect_equal(ntr$Transitive, expected = 0, tolerance = 0)
    expect_equal(ntr$NonTransitive, expected = 0, tolerance = 0)
    expect_equal(ntr$NTR1, expected = 0, tolerance = 0)
    expect_equal(ntr$NTR2, expected = NaN, tolerance = 0)
    
    numTests <- numTests + 1
  }
  
  #-----------------------------------
  #test: complete matrices
  #-----------------------------------
  for(N in 5:20)
  {
    A <- matrix(rep(1, N*N), nrow = N, ncol = N)
    for(i in 1:N) A[i,i] <- 1
    
    ntr <- NTR(A)
    expect_equal(ntr$NumRelations, expected = choose(N,3), tolerance = 0)
    expect_equal(ntr$Transitive, expected = choose(N,3), tolerance = 0)
    expect_equal(ntr$NonTransitive, expected = 0, tolerance = 0)
    expect_equal(ntr$NTR1, expected = 0, tolerance = 0)
    expect_equal(ntr$NTR2, expected = 0, tolerance = 0)
    
    numTests <- numTests + 1
  }
  
  #-----------------------------------
  #test: transitive matrices
  #-----------------------------------
  for(N in 10:20)
    for(trial in 1:10)
    {
      A <- matrix(rep(0, N*N), nrow = N, ncol = N)
      for(i in 1:N) A[i,i] <- 1
      
      cc <- combn(N,4)
      array <- sort(cc[,sample(1:ncol(cc), 1)])
      a <- array[[1]]
      b <- array[[2]]
      c <- array[[3]]
      d <- array[[4]]
      A[a,b] <- 1
      A[b,a] <- 1
      A[a,c] <- 1
      A[c,a] <- 1
      A[b,c] <- 1
      A[c,b] <- 1
      
      ntr <- NTR(A)
      expect_equal(ntr$NumRelations, expected = choose(N,3), tolerance = 0)
      expect_equal(ntr$Transitive, expected = 1, tolerance = 0)
      expect_equal(ntr$NonTransitive, expected = 0, tolerance = 0)
      expect_equal(ntr$NTR1, expected = 0, tolerance = 0)
      expect_equal(ntr$NTR2, expected = 0, tolerance = 0)
      numTests <- numTests + 1
      
      A[a,d] <- 1
      A[d,a] <- 1
      ntr <- NTR(A)
      expect_equal(ntr$Transitive, expected = 1, tolerance = 0)
      expect_equal(ntr$NonTransitive, expected = 2, tolerance = 0)
      expect_equal(ntr$NTR1, expected = 2/choose(N,3)*100, tolerance = 0)
      expect_equal(ntr$NTR2, expected = 2/3*100, tolerance = 0)
      numTests <- numTests + 1
      
      A[c,d] <- 1
      A[d,c] <- 1
      ntr <- NTR(A)
      expect_equal(ntr$Transitive, expected = 2, tolerance = 0)
      expect_equal(ntr$NonTransitive, expected = 2, tolerance = 0)
      expect_equal(ntr$NTR1, expected = 2/choose(N,3)*100, tolerance = 0)
      expect_equal(ntr$NTR2, expected = 2/4*100, tolerance = 0)    
      numTests <- numTests + 1
      
      A[b,d] <- 1
      A[d,b] <- 1
      ntr <- NTR(A)
      expect_equal(ntr$Transitive, expected = 4, tolerance = 0)
      expect_equal(ntr$NonTransitive, expected = 0, tolerance = 0)
      expect_equal(ntr$NTR1, expected = 0, tolerance = 0)
      expect_equal(ntr$NTR2, expected = 0, tolerance = 0)    
      numTests <- numTests + 1
    }
  
  #-----------------------------------
  #test: transitive matrices
  #-----------------------------------
  for(N in 10:20)
  {
    A <- matrix(rep(0, N*N), nrow = N, ncol = N)
    for(i in 1:N) A[i,i] <- 1
    for(i in 1:(N-1))
    {
      A[i,i+1] <- 1
      A[i+1,i] <- 1
    }
    A[N,1] <- 1
    A[1,N] <- 1
    
    ntr <- NTR(A)
    expect_equal(ntr$NumRelations, expected = choose(N,3), tolerance = 0)
    expect_equal(ntr$Transitive, expected = 0, tolerance = 0)
    expect_equal(ntr$NonTransitive, expected = N, tolerance = 0)
    expect_equal(ntr$NTR1, expected = N/choose(N,3)*100, tolerance = 0)
    expect_equal(ntr$NTR2, expected = 100, tolerance = 0)    
    numTests <- numTests + 1
  }
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for Agreement Rates
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.AM.fromDM <- function() 
{
  cat("Running AR tests...")
  numTests <- 0
  
  for(N in seq(5, 100, 5)) # try out samples of different sizes N
    for(repetition in 1:10)
    {
      DM <- generateDM.Simple(N)$DM
      for(epsilon in seq(0.0, 1.0, 0.05)) # try out different tolerance levels
      {
        numTests <- numTests + 1
  
        A <- AM.fromDM(DM, epsilon)
        
        # test: A has the same number of rows and columns as the dissimilarity matrix DM
        expect_equal(nrow(A), expected = N)
        expect_equal(ncol(A), expected = N)
        
        # test: A is symmetric
        expect_true(isSymmetric(A))
        
        # test: A contains values 0 and 1 only
        unique.values <- sort(unique(as.vector(A)))
        if (length(unique.values) == 2) expect_true(all(unique.values == c(0,1)))
        if (length(unique.values) == 1) expect_true(any(unique.values[[1]] == 0, unique.values[[1]] == 1))
        
        # test: A has unit values on the first diagonal
        expect_true(all(unique(diag(A)) == c(1)))
      }
    }
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.Lin.distance <- function() 
{
  cat("Running Lin distance tests...")
  numTests <- 0
  
  #---------------------------------------
  #test: identical arrays, identical probabilities
  #---------------------------------------
  for(N in 1:100)
  {
    a <- rep(N/2, N)
    b <- rep(N/2, N)
    prob <- rep(0.25, N)
    expect_equal(Lin.distance(a, b, prob, prob), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: identical arrays, different probabilities
  #---------------------------------------
  for(N in 1:100)
  {
    a <- rep(N/2, N)
    b <- rep(N/2, N)
    prob <- runif(N, 0, 1)
    expect_equal(Lin.distance(a, b, prob, prob), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: different arrays, identical probabilities
  #---------------------------------------
  for(N in 5:100)
  {
    a <- rep(0, N)
    b <- rep(0, N)
    M <- round(N/3)
    indexes <- sample(1:N, 2*M, replace = FALSE)
    a[indexes[1:M]] <- 1
    b[indexes[(M+1):(2*M)]] <- 1
    prob <- rep(0.35, N)
    expect_equal(Lin.distance(a, b, prob, prob), expected = -((N-2*M)*log(prob[[1]]) + 2*M*log(2*prob[[1]]))/(N*log(prob[[1]])), tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: different arrays, different probabilities
  #---------------------------------------
  for(N in 5:100)
  {
    a <- rep(0, N)
    b <- rep(0, N)
    M <- round(N/3)
    indexes <- sample(1:N, 2*M, replace = FALSE)
    a[indexes[1:M]] <- 1
    b[indexes[(M+1):(2*M)]] <- 1
    prob <- runif(N, 0, 1)
    
    w <- sum(2*log(prob))
    S <- 0
    for(i in 1:N)
      S <- S + ifelse(a[[i]]==b[[i]], log(prob[[i]]), log(2*prob[[i]]))

    expect_equal(Lin.distance(a, b, prob, prob), expected = -2*S/w, tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #custom tests
  #---------------------------------------
  expect_equal(Lin.distance(c(4,1,7,3,8), c(6,1,5,2,8), c(0.4,0.1,0.7,0.3,0.8), c(0.6,0.1,0.5,0.2,0.8)), expected = -0.5872387, tolerance = 1e-3)
  numTests <- numTests + 1
  
  expect_equal(Lin.distance(c(4,1,7,3,8), c(6,1,5,2,8), c(0.2,0.2,0.2,0.2,0.2), c(0.2,0.2,0.2,0.2,0.2)), expected = -0.7415941, tolerance = 1e-3)
  numTests <- numTests + 1
  
  expect_equal(Lin.distance(c(7,1,5,8,2,6,3,4), c(8,1,6,3,2,3,5,4), c(0.7,0.1,0.5,0.8,0.2,0.6,0.3,0.4), c(0.8,0.1,0.6,0.3,0.2,0.3,0.5,0.4)), expected = -0.5535053, tolerance = 1e-3)
  numTests <- numTests + 1
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.Lin.distance.ZM <- function() 
{
  cat("Running Lin distance tests...")
  numTests <- 0
  
  #---------------------------------------
  #test: identical arrays, identical ZM models
  #---------------------------------------
  ZM <- lnre("zm", alpha = 1/2, B = 0.62)
  for(N in 1:100)
  {
    a <- rep(1, N)
    b <- rep(1, N)
    ZModels <- list()
    for(i in 1:N) ZModels[[i]] <- ZM
    expect_equal(Lin.distance.ZM(a, b, ZModels), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: different arrays, identical ZM models
  #---------------------------------------
  ZM <- lnre("zm", alpha = 1/2, B = 0.62)
  for(N in 1:100)
  {
    a <- rep(1, N)
    b <- rep(2, N)
    ZModels <- list()
    for(i in 1:N) ZModels[[i]] <- ZM
    expect_equal(Lin.distance.ZM(a, b, ZModels), expected = -2*log(0.382716+0.1708554)/(log(0.382716)+log(0.1708554)), tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: different arrays, identical ZM models
  #---------------------------------------
  ZM <- lnre("zm", alpha = 1/2, B = 0.62)
  for(N in 1:100)
  {
    a <- rep(1, N)
    b <- rep(1, N)
    M <- round(N/3)
    indexes <- sample(1:N, 2*M, replace = FALSE)
    a[indexes[1:M]] <- 2
    b[indexes[(M+1):(2*M)]] <- 3
    
    ZModels <- list()
    for(i in 1:N) ZModels[[i]] <- ZM
    expect_equal(Lin.distance.ZM(a, b, ZModels), expected = -2*((N-2*M)*log(0.382716) + M*log(0.382716+0.1708554) + M*log(0.382716+0.09677822))/(2*(N-M)*log(0.382716)+M*log(0.1708554)+M*log(0.09677822)), tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #custom tests
  #---------------------------------------
  ZM1 <- lnre("zm", alpha = 1/2, B = 0.62)
  ZM2 <- lnre("zm", alpha = 1/2, B = 2.25)
  expect_equal(Lin.distance.ZM(c(1,2),c(3,4),list(ZM1,ZM2)), expected = -0.5837606, tolerance = 1e-3)
  numTests <- numTests + 1
  
  expect_equal(Lin.distance.ZM(c(1,7,1,4),c(2,3,1,8),list(ZM1,ZM2,ZM2,ZM1)), expected = -0.7014183, tolerance = 1e-3)
  numTests <- numTests + 1
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.Lin.distance.Bernoulli <- function() 
{
  cat("Running Lin distance tests...")
  numTests <- 0

  #---------------------------------------
  #test: identical arrays, identical probabilities
  #---------------------------------------
  for(N in 1:100)
  {
    a <- rep(0, N)
    b <- rep(0, N)
    p <- rep(0.5, N)
    expect_equal(Lin.distance.Bernoulli(a, b, p), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
    
    a <- rep(1, N)
    b <- rep(1, N)
    expect_equal(Lin.distance.Bernoulli(a, b, p), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
  }

  #---------------------------------------
  #test: identical arrays, different probabilities
  #---------------------------------------
  for(N in 1:100)
  {
    a <- rep(0, N)
    b <- rep(0, N)
    p <- runif(N, 0, 1)
    expect_equal(Lin.distance.Bernoulli(a, b, p), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
    
    a <- rep(1, N)
    b <- rep(1, N)
    expect_equal(Lin.distance.Bernoulli(a, b, p), expected = -1, tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #test: different arrays, identical probabilities
  #---------------------------------------
  for(N in 5:100)
  {
    a <- rep(0, N)
    b <- rep(0, N)
    M <- round(N/3)
    indexes <- sample(1:N, 2*M, replace = FALSE)
    a[indexes[1:M]] <- 1
    b[indexes[(M+1):(2*M)]] <- 1
    
    p <- rep(0.5, N)
    expect_equal(Lin.distance.Bernoulli(a, b, p), expected = -(2*(N-2*M)*log(1-p[[1]]))/( 2*M*log(p[[1]]) + 2*(N-M)*log(1-p[[1]]) ), tolerance = 1e-3)
    numTests <- numTests + 1
  }
  
  #---------------------------------------
  #custom tests
  #---------------------------------------
  expect_equal(Lin.distance.Bernoulli(c(0,1,0,0,1,0),c(1,1,1,1,0,0),c(0.2,0.3,0.4,0.5,0.6,0.1)), expected = -0.301281, tolerance = 1e-3)
  numTests <- numTests + 1
  
  expect_equal(Lin.distance.Bernoulli(c(0,1,0,0,0,1,0,1,0,1),c(1,1,1,1,0,1,1,1,0,0),c(0.7,0.3,0.2,0.4,0.5,0.4,0.3,0.2,0.1,0.8)), expected = -0.524402, tolerance = 1e-3)
  numTests <- numTests + 1
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.overlap.distance <- function() 
{
  cat("Running Overlap distance tests...")
  numTests <- 0
  
  ##------------------------------------
  # test: zero distance
  ##------------------------------------
  for(N in 1:100) # various sizes N of the two sets that are compared
  {
    a <- rep("test", N)
    b <- rep("test", N)
    expect_equal(overlap.distance(a,b), expected = 0, tolerance = 0)
    numTests <- numTests + 1
  }
  
  ##------------------------------------
  # test: maximum distance (1.0)
  ##------------------------------------
  for(N in 1:100) # various sizes N of the two sets that are compared
  {
    a <- sample(c(1,100), N, replace = TRUE)
    b <- sample(c(101,200), N, replace = TRUE)
    expect_equal(overlap.distance(a,b), expected = 1, tolerance = 0)
    numTests <- numTests + 1
  }
  
  ##------------------------------------
  # test: various distances between 0.0 and 1.0
  ##------------------------------------
  for(N in 1:100) # various sizes of the two sets that are compared
    for(matchCount in 1:N)
    {
      a <- rep("a", N) 
      b <- rep("b", N)
      indexes <- sample(1:N, matchCount, replace = FALSE)
      for(i in 1:length(indexes))
      {
        a[[indexes[[i]]]] <- i
        b[[indexes[[i]]]] <- i
      }
      expect_equal(overlap.distance(a,b), expected = 1 - matchCount/N, tolerance = 1e-6)
      numTests <- numTests + 1
    }
  
  ##------------------------------------
  # custom tests:
  ##------------------------------------
  expect_equal(overlap.distance(c(1,2,3,4,5,6), c(1,7,3,8,5,9)), expected = 3/6, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(1,2,3,4,5,6,10,11,12), c(1,7,3,8,5,9,10,11,12)), expected = 3/9, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(-1,0,2,1,2,3,4,5,6,10,11,12), c(-1,3,3,1,7,3,8,5,9,10,11,12)), expected = 5/12, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(0,1,2,3,4,5,6,7,8,9,10), c(2,1,2,3,5,5,5,5,5,5,5)), expected = 7/11, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(1), c(2)), expected = 1.0, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(1), c(1)), expected = 0.0, tolerance = 0)
  numTests <- numTests + 1
  
  expect_equal(overlap.distance(c(7,1,5,8,3,8,4,6,9), c(10,2,4,8,8,8,8,6,10)), expected = 6/9, tolerance = 0)
  numTests <- numTests + 1
  
  cat(numTests, "done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.Simple <- function() 
{
  cat("Running Dissimilarity Matrix tests (Manhattan) ...\r\n")
  numErrors <- 0
  means <- c()
  SDs <- c()
  
  TRIALS <- 1000
  for(trial in 1:TRIALS)
  {
    data <- generateDM.Simple(N = 100, c(0, 1))
    
    # test: matrix should be symmetric
    expect_true(isSymmetric(data$DM))
    
    # test: matrix values should be 0 on the main diagonal
    expect_true(all(unique(diag(data$DM)) == c(0)))
    
    # test: values should be normally distributed
    if (shapiro.test(data$values)$p.value < 0.05) 
      numErrors <- numErrors + 1
    
    # means should be 0 and SDs should be 1
    means[[trial]] <- mean(data$values)
    SDs[[trial]] <- sd(data$values)
  }
  
  cat("Mean =", mean(means), "\r\nSD =", mean(SDs), "\r\n")
  cat("Shapiro-Wilk normality test was significant in ", numErrors, " of ", TRIALS, " tests (", sprintf("%.2f", numErrors/TRIALS),").\r\n", sep = "")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.Minkowski <- function() 
{
  cat("Running Dissimilarity Matrix tests (Minkowski) ...")
  numTests <- 0
  
  for(N in seq(10,50,10))
    for(k in 3:10)
    {
      # test: Minkowski with p = 2 should be identical with the Euclidean distance
      data <- generateDM.Minkowski(N, params = seq(1,k,1), p = 2)
      expect_true(isSymmetric(data$DM))
      expect_true(all(unique(diag(data$DM)) == c(0)))
      for(i in 1:N)
        for(j in i:N)
          expect_equal(data$DM[i,j], expected = dist(rbind(data$values[i,], data$values[j,]))[[1]], tolerance = 1e-5)
      numTests <- numTests + 1
      
      # test: only proposals are required
      data <- generateDM.Minkowski(N, params = seq(1,k,1), p=2, proposals = NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      expect_false(is.null(data$values))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- generateDM.Minkowski(N, params = seq(1,k,1), p=2, proposals = data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- generateDM.Minkowski(N, params = seq(1,k,1), p=2, proposals = data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
      
      # test: Minkowski with p = 1 should be identical with the Manhattan distance
      data <- generateDM.Minkowski(N, params = seq(1,k,1), p = 1)
      expect_true(isSymmetric(data$DM))
      expect_true(all(unique(diag(data$DM)) == c(0)))
      for(i in 1:N)
        for(j in i:N)
          expect_equal(data$DM[i,j], expected = sum(abs(data$values[i,] - data$values[j,])), tolerance = 1e-5)
      numTests <- numTests + 1
    }
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.SqrEuclidean <- function() 
{
  cat("Running Dissimilarity Matrix tests (Sqr Euclidean) ...")
  numTests <- 0
  
  for(N in seq(10,50,10))
    for(k in 3:10)
    {
      data <- generateDM.SqrEuclidean(N, params = seq(1,k,1))
      expect_true(isSymmetric(data$DM))
      expect_true(all(unique(diag(data$DM)) == c(0)))
      for(i in 1:N)
        for(j in i:N)
          expect_equal(data$DM[i,j], expected = dist(rbind(data$values[i,], data$values[j,]))[[1]]^2, tolerance = 1e-5)
      numTests <- numTests + 1
      
      # test: only proposals are required
      data <- generateDM.SqrEuclidean(N, params = seq(1,k,1), proposals = NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      expect_false(is.null(data$values))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- generateDM.SqrEuclidean(N, params = seq(1,k,1), proposals = data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- generateDM.SqrEuclidean(N, params = seq(1,k,1), proposals = data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
    }
  
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.ZipfMandelbrot.Overlap <- function() 
{
  cat("Running Dissimilarity Matrix tests (Overlap) ...")
  numTests <- 0
  for(N in seq(10,100,10))
    for(k in 3:10)
    {
      data <- generateDM.ZipfMandelbrot.Overlap(N, runif(k))
      
      # test: matrix should be symmetric
      expect_true(isSymmetric(data$DM))
      
      # test: matrix values should be 0 on the main diagonal
      expect_true(all(unique(diag(data$DM)) == c(0)))
      numTests <- numTests + 1
      
      # test: only proposals are required
      ps <- runif(k)
      data <- generateDM.ZipfMandelbrot.Overlap(N, ps, NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      expect_false(is.null(data$values))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- generateDM.ZipfMandelbrot.Overlap(N, ps, data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- generateDM.ZipfMandelbrot.Overlap(N, ps, data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
    }
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.ZipfMandelbrot.Lin <- function() 
{
  cat("Running Dissimilarity Matrix tests (Lin) ...")
  numTests <- 0
  for(N in seq(10,50,10))
    for(k in 3:10)
    {
      data <- generateDM.ZipfMandelbrot.Lin(N, runif(k))
      
      # test: matrix should be symmetric
      expect_true(isSymmetric(data$DM))
      
      # test: matrix values should be 0 on the main diagonal
      expect_true(all(unique(diag(data$DM)) == c(0)))
      numTests <- numTests + 1
      
      # test: only proposals are required
      ps <- runif(k)
      data <- generateDM.ZipfMandelbrot.Lin(N, ps, proposals = NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      expect_false(is.null(data$values))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- generateDM.ZipfMandelbrot.Lin(N, ps, proposals = data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- generateDM.ZipfMandelbrot.Lin(N, ps, proposals = data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
    }
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM.Bernoulli.Lin <- function() 
{
  cat("Running Dissimilarity Matrix tests (Lin) ...")
  numTests <- 0
  for(N in seq(10,100,10))
    for(k in 3:10)
    {
      data <- generateDM.Bernoulli.Lin(N, runif(k))
      
      # test: matrix should be symmetric
      expect_true(isSymmetric(data$DM))
      
      # test: matrix values should be 0 on the main diagonal
      expect_true(all(unique(diag(data$DM)) == c(0)))
      numTests <- numTests + 1
      
      # test: only proposals are required
      ps <- runif(k)
      data <- generateDM.Bernoulli.Lin(N, ps, proposals = NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      expect_false(is.null(data$values))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- generateDM.Bernoulli.Lin(N, ps, proposals = data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- generateDM.Bernoulli.Lin(N, ps, proposals = data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
    }
  cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Unit tests for dissimilarity functions, distances, and dissimilarity matrices
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.generateDM <- function(FUN, params, TRIALS = 10, TRIANGLE_INEQUALITY = TRUE, VERBOSE = FALSE) 
{
  if (VERBOSE) cat("Running Dissimilarity Matrix tests ...")
  numTests <- 0
  
  for(N in seq(10, 100, 10)) # test with samples of different sizes
  {
    if (VERBOSE) cat("Running tests for N =", N, "\r\n")
    
    for(trial in 1:TRIALS) # multiple trials for each N
    {
      # generate dissimilarity matrix
      data <- FUN(N, params)
      DM <- data$DM
      
      # test: the number of rows and columns should be N
      expect_equal(nrow(DM), expected = N)
      expect_equal(ncol(DM), expected = N)
      numTests <- numTests + 2
      
      # test: all the values should be positive
      expect_gte(min(DM), 0)
      numTests <- numTests + 1
      
      # test: all the values of the first diagonal should be 0
      values <- unique(diag(DM))
      expect_equal(length(values), expected = 1, tolerance = 0)
      expect_equal(values[[1]], 0, tolerance = 0)
      numTests <- numTests + 2
      
      # test: DM should be symmetric
      expect_true(isSymmetric(DM))
      numTests <- numTests + 1
      
      # test: DM should respect the triangle inequality only if computed from a metric
      tri <- unittests.triangleInequality(DM)
      if (any(tri == FALSE, TRIANGLE_INEQUALITY == TRUE)) # triangle inequality can occur, even if not expected
        expect_equal(tri, TRIANGLE_INEQUALITY) 
      numTests <- numTests + 1
      
      # test: only proposals are required
      data <- FUN(N, params, proposals = NULL, GENERATE_PROPOSALS_ONLY = TRUE)
      expect_true(is.null(data$DM))
      numTests <- numTests + 1
      
      # test: predefined proposals
      datanew <- FUN(N, params, proposals = data$values)
      expect_equal(datanew$values, data$values)
      numTests <- numTests + 1
      
      # test: predefined proposals (2)
      datanew.second <- FUN(N, params, proposals = data$values)
      expect_equal(datanew.second$values, datanew$values)
      expect_equal(datanew.second$DM, datanew$DM)
      numTests <- numTests + 1
    }
  }
  
  if (VERBOSE) cat(numTests, "tests done.\r\n")
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Adapted from https://rdrr.io/cran/fossil/src/R/tri.ineq.R by adding a tolerance when comparing real values
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unittests.triangleInequality <- function(dist, tolerance = 1e-10) 
{
  mat<-as.matrix(dist)
  n<-dim(mat)[1]
  ineq <- 0
  for (i in 1:(n-2)) {
    for (j in (i+1):(n-1)) {
      for (k in (j+1):n) {
        sds<-c(mat[j,i], mat[k,i], mat[k,j])
        lng<-max(sds)
        if (lng>(sum(sds)-lng+tolerance)) 
          return (FALSE)
      }
    }
  }
  return(TRUE)
}
